local VERSION = tonumber(("$Revision: 421 $"):match("%d+"))

local QuestsFu = QuestsFu
local QuestsFu_Tracker = QuestsFu:NewModule("Tracker", "AceEvent-2.0", "AceHook-2.1", "AceConsole-2.0")
if QuestsFu.revision < VERSION then
	QuestsFu.version = "r" .. VERSION
	QuestsFu.revision = VERSION
	QuestsFu.date = ("$Date: 2009-12-15 02:24:09 +0000 (Tue, 15 Dec 2009) $"):match("%d%d%d%d%-%d%d%-%d%d")
end

local tablet = AceLibrary("Tablet-2.0")
local quixote = LibStub("LibQuixote-2.0")
local dewdrop = AceLibrary("Dewdrop-2.0")

local L = AceLibrary("AceLocale-2.2"):new("QuestsFu_Tracker")
local Lcore = AceLibrary("AceLocale-2.2"):new("QuestsFu")

QuestsFu_Tracker.lname = L["Tracker"]
QuestsFu_Tracker.desc = L["Replace the default quest tracker with a slightly more featureful one"]

local debug = function(...)
	--QuestsFu_Tracker:PrintLiteral(...)
end

function QuestsFu_Tracker:OnInitialize()
	self.db = QuestsFu:AcquireDBNamespace("Tracker")
	QuestsFu:RegisterDefaults("Tracker", "profile", {
		login = true,
		zone = false,
		subzone = true,
		gained = false,
		progress = true,
		removeCompleted = false,
		useOwnTracker = true,
		data = {
			fontSizePercent=0.7, detached=true, transparency=0,
		},
		strata = 'LOW',
		minWidth = 150,
		maxHeight = 800,
	})
	QuestsFu:RegisterDefaults("Tracker", "char", {
		watchedQuests = {},
		autowatched = {},
	})
	
	self.menu = {
		name = L["Tracker"],
		desc = L["Replace the default quest tracker with a slightly more featureful one"],
		type = "group",
		args = {
			auto = {
				name = L["Autowatch"],
				desc = L["Automatically add quests to the tracker when appropriate"],
				type = "group",
				args = {
					login = {
						type = "toggle", name = L["Login"],
						desc = L["Re-add quests you were watching before you logged out"],
						get = function() return self.db.profile.login end,
						set = function(t) self.db.profile.login = t end,
					},
					zone = {
						type = "toggle", name = L["Zone"],
						desc = L["Add quests to the tracker when you enter their zone"],
						get = function() return self.db.profile.zone end,
						set = function(t) self.db.profile.zone = t end,
					},
					subzone = {
						type = "toggle", name = L["Subzone"],
						desc = L["Add quests to the tracker when you enter their subzone"],
						get = function() return self.db.profile.subzone end,
						set = function(t) self.db.profile.subzone = t end,
					},
					progress = {
						type = "toggle", name = L["Progress"],
						desc = L["Add quests to the tracker when you make progress on them"],
						get = function() return self.db.profile.progress end,
						set = function(t) self.db.profile.progress = t end,
					},
					gained = {
						type = "toggle", name = L["Gained"],
						desc = L["Add quests to the tracker when they are first gained"],
						get = function() return self.db.profile.gained end,
						set = function(t) self.db.profile.gained = t end,
					},
					removeCompleted = {
						type = "toggle", name = L["Remove completed"],
						desc = L["Remove quests from the tracker when you complete their objectives"],
						get = function() return self.db.profile.removeCompleted end,
						set = function(t) self.db.profile.removeCompleted = t end,
					},
				},
			},
			own = {
				type = 'toggle', name = L["Use own tracker"],
				desc = L["Replace the default quest tracker with a slightly more featureful one"],
				get = function() return self.db.profile.useOwnTracker end,
				set = function(t)
					self.db.profile.useOwnTracker = t
					if self.db.profile.useOwnTracker then
						if tablet:IsRegistered('QuestsFu_Tracker') then tablet:Open('QuestsFu_Tracker') end
						self:Update()
					else
						if tablet:IsRegistered('QuestsFu_Tracker') then tablet:Close('QuestsFu_Tracker') end
						WatchFrame:Show()
						--WatchFrameTitleButton:Show()
						WatchFrameLines:Show()
					end
				end,
			},
			strata = QuestsFu.strataOption(self, 'QuestsFu_Tracker', 1),
			maxHeight = QuestsFu.maxHeightOption(self, 'QuestsFu_Tracker', 2),
			minWidth = QuestsFu.minWidthOption(self, 'QuestsFu_Tracker', 3),
			--This one will look a little weird on the tablet right-click menu, as it'll be duplicating native functionality.  But it's totally worth it to stop people complaining.
			lock = QuestsFu.lockOption(self, 'QuestsFu_Tracker', 4),
		},
	}
end

function QuestsFu_Tracker:OnEnable()
	if quixote.firstDone then
		self:SecureHook("AddQuestWatch")
		self:SecureHook("RemoveQuestWatch")
		self:RegisterEvent("MINIMAP_ZONE_CHANGED", "OnZoneChange")
		self:RegisterEvent("ZONE_CHANGED", "OnZoneChange")
		self:RegisterEvent("ZONE_CHANGED_INDOORS", "OnZoneChange")
		self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "OnZoneChange")
		quixote.RegisterCallback(self, 'Objective_Update')
		quixote.RegisterCallback(self, 'Quest_Gained')
		quixote.RegisterCallback(self, 'Quest_Complete')
		quixote.RegisterCallback(self, 'Quest_Lost')
		self:Hook(QuestsFu, "ToggleQuestWatch")
		self:Hook(QuestsFu, "IsWatched")
		
		self:Hook(QuestsFu, "UpdateTooltip")
		self:SecureHook("WatchFrame_Update", "Update")
		quixote.RegisterCallback(self, 'Update')
		quixote.RegisterCallback(self, 'Party_Update', 'Update')
		self:RegisterEvent("TRACKED_ACHIEVEMENT_UPDATE", "Update")

		--Check various bits of saved info for continuing validity
		self:LoadWatchedQuests()
		self:OnZoneChange()

		self.oldMAX_WATCHABLE_QUESTS = MAX_WATCHABLE_QUESTS
		MAX_WATCHABLE_QUESTS = 50
	else
		quixote.RegisterCallback(self, "Ready", function() self:OnEnable() end)
	end
end

function QuestsFu_Tracker:OnDisable()
	MAX_WATCHABLE_QUESTS = self.oldMAX_WATCHABLE_QUESTS
	QuestsFu.toggleTablet('QuestsFu_Tracker', false)
	tablet:Unregister('QuestsFu_Tracker')
end

--- Make other QuestsFu modules deal with this ---

function QuestsFu_Tracker:ToggleQuestWatch(object, questid)
	if self.db.profile.useOwnTracker then
		if self.db.char.watchedQuests[questid] then
			self.db.char.watchedQuests[questid] = nil
		else
			self.db.char.watchedQuests[questid] = true
		end
		self:Update()
	else
		return self.hooks[object].ToggleQuestWatch(object, questid)
	end
end

function QuestsFu_Tracker:IsWatched(object, uid)
	if type(uid) == 'string' then
		--need the uid
		uid = quixote:GetQuest(uid)
	end
	return self.db.char.watchedQuests[uid]
	--not calling of the hook
end

--- Quest watching ---

function QuestsFu_Tracker:LoadWatchedQuests()
	for uid in pairs(self.db.char.watchedQuests) do
		if not quixote:GetQuestByUid(uid) then
			self.db.char.watchedQuests[uid] = nil
		end
	end
	for i,uid in pairs(self.db.char.autowatched) do
		if not quixote:GetQuestByUid(uid) then
			table.remove(self.db.char.autowatched, i)
		end
	end
	if not self.db.profile.useOwnTracker then
		if GetNumQuestWatches() == 0 then
			for uid in pairs(self.db.char.watchedQuests) do
				local _, questId, title = quixote:GetQuestByUid(uid)
				if questId then
					AddQuestWatch(questId)
				else
					self.db.char.watchedQuests[uid] = nil
				end
			end
		end
	end
	self.loadedWatchedQuests = true
	self:Update()
end

function QuestsFu_Tracker:SaveWatchedQuests()
	if self.db.profile.useOwnTracker then
		self:Update()
	else
		if self.loadedWatchedQuests then
			table.wipe(self.db.char.watchedQuests)
			for i = 1, GetNumQuestWatches() do
				local uid = quixote:GetQuestById(GetQuestIndexForWatch(i))
				if uid then self.db.char.watchedQuests[uid] = true end
			end
		end
	end
end

function QuestsFu_Tracker:AddQuestWatch(questid)
	local uid, _, title = quixote:GetQuestById(questid)
	if uid then
		self.db.char.watchedQuests[uid] = true
	end
	debug("AddQuestWatch", questid, uid, title)
	if self.db.profile.useOwnTracker then
		WatchFrame:Hide()
		--WatchFrameTitleButton:Hide()
		WatchFrameLines:Hide()
	end
end

function QuestsFu_Tracker:RemoveQuestWatch(questid)
	local uid, _, title = quixote:GetQuestById(questid)
	if uid then
		self.db.char.watchedQuests[uid] = nil
	end
	for i,wuid in pairs(self.db.char.autowatched) do
		if wuid == uid then
			table.remove(self.db.char.autowatched, i)
			break
		end
	end
	debug("RemoveQuestWatch", questid, uid, title)
end

function QuestsFu_Tracker:Quest_Lost(event, name, uid, zone)
	self.db.char.watchedQuests[uid] = nil
	if #self.db.char.autowatched > 0 then
		for i,wuid in pairs(self.db.char.autowatched) do
			if uid == wuid then
				table.remove(self.db.char.autowatched, i)
				break
			end
		end
	end
end

function QuestsFu_Tracker:OnZoneChange()
	debug("OnZoneChange")
	if UnitIsDeadOrGhost('player') then return end
	--Get rid of autowatched quests
	for i, uid in pairs(self.db.char.autowatched) do
		local _,questid,title = quixote:GetQuestByUid(uid)
		if questid then
			debug("Removing autowatched", uid, questid, title)
			RemoveQuestWatch(questid)
		end
	end
	table.wipe(self.db.char.autowatched)
	if not (self.db.profile.zone or self.db.profile.subzone) then return end
	local zone = GetRealZoneText()
	local subzone = GetSubZoneText()
	for _, uid, questid, name, _, _, _, complete in quixote:IterateQuestsInZone(zone) do
		-- If we're already watching it, don't autowatch.  (This way it won't be de-watched when we leave the zone.)
		if not (complete and self.db.profile.removeCompleted) and not self.db.char.watchedQuests[uid] then
			if self.db.profile.zone then
				debug("Autowatching", "zone", uid, questid, name)
				table.insert(self.db.char.autowatched, uid)
				AddQuestWatch(questid)
			elseif self.db.profile.subzone and subzone ~= "" then
				local description, objective = quixote:GetQuestText(uid)
				if objective:find(subzone, 1, true) or description:find(subzone, 1, true) then
					debug("Autowatching", "subzone", uid, questid, name)
					table.insert(self.db.char.autowatched, uid)
					AddQuestWatch(questid)
				end
			end
		end
	end
	self:Update()
end

function QuestsFu_Tracker:Objective_Update(event, name, uid, desc, numHad, numGot, numNeeded)
	if AUTO_QUEST_WATCH=="0" and self.db.profile.progress then
		debug("Watching for progress", uid, name)
		local _,id = quixote:GetQuestByUid(uid)
		AddQuestWatch(id)
	end
	self:Update()
end

function QuestsFu_Tracker:Quest_Gained(event, name, uid)
	local _,id = quixote:GetQuestByUid(uid)
	if self.db.profile.gained then
		debug("Watching for gained", uid, name)
		--table.insert(self.db.char.autowatched, uid)
		AddQuestWatch(id)
	elseif self.db.profile.zone then
		local zone = GetRealZoneText()
		local questzone = select(10, quixote:GetQuestByUid(uid))
		if zone == questzone then
			debug("Watching for zone", uid, name)
			table.insert(self.db.char.autowatched, uid)
			AddQuestWatch(id)
		end
	elseif self.db.profile.subzone then
		local subzone = GetSubZoneText()
		local description, objective = quixote:GetQuestText(uid)
		if subzone ~= "" and (objective:find(subzone, 1, true) or description:find(subzone, 1, true)) then
			debug("Watching for subzone", uid, name)
			table.insert(self.db.char.autowatched, uid)
			AddQuestWatch(id)
		end
	end
	self:Update()
end

function QuestsFu_Tracker:Quest_Complete(event, name, uid)
	if self.db.profile.removeCompleted then
		local _,id = quixote:GetQuestByUid(uid)
		RemoveQuestWatch(id)
	end
	self:Update()
end

--- Tracker ---

function QuestsFu_Tracker:UpdateTooltip(object)
	self:Update()
	return self.hooks[object].UpdateTooltip(object)
end

function QuestsFu_Tracker:AddAchievementsToCategory(cat, ...)
	local numAchievements = select('#', ...)
	if numAchievements == 0 then
		return
	end
	for i=1, numAchievements do
		local aid, name, points, completed, month, day, year, description, flags, image, rewardtext = GetAchievementInfo((select(i, ...)))
		if QuestsFu.db.profile.showLevelsTablet then
			name = '[A'..tostring(points)..'] '..name
		end
		if not QuestsFu.db.profile.color.title then name = QuestsFu:Colorize(name, 1, 1, 1) end
		cat:AddLine(
			'text', name, 'wrap', self.db.profile.wrapQuests,
			'text2', completed and Lcore["(done)"] or '',
			'textR', completed and 0 or 1, 'textG', 1, 'textB', completed and 0 or 1,
			'text2R', completed and 0 or 1, 'text2G', 1, 'text2B', completed and 0 or 1,
			'indentation', 6,
			'func', false -- overrides the category default
		)
		local num_criteria = GetAchievementNumCriteria(aid)
		if num_criteria == 0 then
			if not (completed and not QuestsFu.db.profile.showCompletedObjectives) then
				cat:AddLine(
					'text', description, 'wrap', true,
					'textR', completed and 0 or 1, 'textG', 1, 'textB', completed and 0 or 1,
					'size', tablet:GetNormalFontSize()-2,
					'indentation', 12,
					'func', false
				)
			end
		else
			for ii=1, num_criteria do
				local critera, ctype, ccompleted, quantity, reqquantity, charname, flags, assetID, quantitystring, cid = GetAchievementCriteriaInfo(aid, ii)
				if not (ccompleted and not QuestsFu.db.profile.showCompletedObjectives) then
					local r,g,b,a
					if QuestsFu.db.profile.color.objectives then
						r,g,b,a = QuestsFu:GetColorFromCompletion(quantity / reqquantity)
					end
					cat:AddLine(
						'text', '  '..critera,
						'text2', ccompleted and Lcore["(done)"] or string.format("%s/%s", quantity, reqquantity),
						'textR', r or 1, 'textG', g or 1, 'textB', b or 1,
						'text2R', r or 1, 'text2G', g or 1, 'text2B', b or 1,
						'size', tablet:GetNormalFontSize()-2, 'size2', tablet:GetNormalFontSize()-2,
						'indentation', 12,
						'func', false
					)
				end
			end
		end
	end
end

do
	local function sortQuestsByLevel(a, b)
		local level_a = select(4, quixote:GetQuestByUid(a))
		local level_b = select(4, quixote:GetQuestByUid(b))
		if not (level_a and level_b) or level_a == level_b then
			return a < b
		else
			return level_a < level_b
		end
	end

	function QuestsFu_Tracker:Update()
		if self.db.profile.useOwnTracker then
			local trackerOrder = {}
			if not tablet:IsRegistered('QuestsFu_Tracker') then
				tablet:Register('QuestsFu_Tracker', 'detachedData', self.db.profile.data, 'cantAttach', true,
					'dontHook', true, 'hideWhenEmpty', true, 'strata', self.db.profile.strata,
					'minWidth', self.db.profile.minWidth, 'maxHeight', self.db.profile.maxHeight,
					'menu', function() dewdrop:FeedAceOptionsTable(self.menu) end,
					'children', function()
						--Sort the quests:
						for uid in pairs(self.db.char.watchedQuests) do
							table.insert(trackerOrder, uid)
						end
						table.sort(trackerOrder, sortQuestsByLevel)
						--Display them:
						tablet:SetTitle("Tracker")
						local cat = tablet:AddCategory('columns', 2, 'child_func', QuestsFu.QuestClick, 'child_arg1', QuestsFu)
						for _,uid in pairs(trackerOrder) do QuestsFu:AddQuestToCategory(cat, uid, false, true, true) end
						--Achievements:
						self:AddAchievementsToCategory(cat, GetTrackedAchievements())
						--Clear up the order table:
						table.wipe(trackerOrder)
					end)
				tablet:Open('QuestsFu_Tracker')
			end
			tablet:Refresh('QuestsFu_Tracker')
			WatchFrame:Hide()
			--WatchFrameTitleButton:Hide()
			WatchFrameLines:Hide()
		end
	end
end
